#!BPY

"""
Name: 'DirectX (.x) [K]...'
Blender: 244
Group: 'Export'
Tip: 'Export to DirectX text file format.'
"""

__author__ = "kakkou"
__url__ = ("t-mse.net", "Home page, http://t-mse.net")
__version__ = "0.82 2007/07/06"
__bpydoc__ = """Exports to Direct3D .x format."""


import Blender
from Blender import Mathutils
from Blender.Mathutils import *
from Blender import Draw, BGL
from Blender.BGL import *

indent_level = 0
Hierarchy_level = 0
AnimTicksPerSecond = 25
AnimTickCtrl = None

IsOutputAll = True
IsSwapYZ = True
IsFlipZ = True
IsAnimetion = True
IsNormals = True
IsVertexDuplication = True
IsMaterials = True
IsRefMaterial = False #Material̊O`[h()
IsTextureCoords = True
IsVertexColors = True
IsSkinMesh = True
IsParentAnimKey = True #IK̂悤ɎqIuWFNĝ݂ɃL[łꍇłe̎pL[ł悤ɂ



contents = []

flipMat = Matrix([1,0,0,0],[0,1,0,0],[0,0,-1,0],[0,0,0,1]) # flipZ
rotMat = Mathutils.RotationMatrix(-90, 4, 'x') # swapYZ


##==================================================##

def sanitizeName( old_name ):
    """ÕTj^CY(̂镶̃GXP[v)"""
    escape_name = old_name.replace('.','_')
    escape_name = escape_name.replace(' ','_')
    if escape_name[0].isdigit():
        escape_name = '_' + escape_name
    return escape_name

def output( string ):
    """Cfgto"""
    global contents, indent_level
#    print indent_level*("\t") + string
    contents += indent_level*("\t") + string + "\n"


#4x4sp[^o
def outputMatrixParam( mat ):
#    output( "x=%f, y=%f, z=%f" % ( mat.toEuler().x, mat.toEuler().y, mat.toEuler().z ) )
    output( "%f,%f,%f,%f," % ( mat[0][0], mat[0][1], mat[0][2], mat[0][3] ) )
    output( "%f,%f,%f,%f," % ( mat[1][0], mat[1][1], mat[1][2], mat[1][3] ) )
    output( "%f,%f,%f,%f," % ( mat[2][0], mat[2][1], mat[2][2], mat[2][3] ) )
    output( "%f,%f,%f,%f;;" % ( mat[3][0], mat[3][1], mat[3][2], mat[3][3] ) )

##==================================================##

#GNX|[^NX
class XFileExporter:
    
    def __init__( self ):
        #UVLłȂ_CfbNXǉĂXg
        self.extension_verts = []
        #_CfbNXYƂUVXg
        self.texcoords_list = []
        #ʃCfbNXYƂg_ӎʃXg(g_̃CfbNX݂͌̑_̌ɒǉ)
        self.face_list = []
        #gςݒ_Xg
        self.vertex_list = []
        #o͂IuWFNgXg
        self.object_list = []

    def getChildren(self,obj):	
	obs = Blender.Scene.GetCurrent().objects
	return [ ob for ob in obs if ob.parent == obj ]
	
    def createVertexList( self, mesh ):
        """_CfbNXɂUVl_Xg̍쐬"""

        #MeshGNX|[g邽тɏ
        self.extension_verts = [] 
        self.texcoords_list = [ [] for i in range( len( mesh.verts) )]
        self.face_list = []
        
        for i, face in enumerate( mesh.faces ):
            face_vert_index = []
            for j, v in enumerate( face.v ):
                #Zɒ_CfbNXĂ
                uv = Vector( face.uv[j][0], face.uv[j][1], v.index )
                shareUV = None
                if len( self.texcoords_list[v.index] ) != 0:
                    #UVLo钸_𒲂ׂ
                    for comp_uv in self.texcoords_list[v.index]:
                        if comp_uv == uv:
                            shareUV = uv
                            break
                    
                    if shareUV:
                        #Loꍇ
                        face_vert_index.append( int( shareUV.z ) )
                    else:
                        #Lo܂ł
                        self.extension_verts.append( v.index )
                        uv.z = len( mesh.verts ) + len( self.extension_verts ) - 1 #_CfbNXgłɍXV
                        self.texcoords_list[v.index].append( uv )
                        face_vert_index.append( int( uv.z ) )
    
                else:
                    #UVXg󂾂ꍇ
                    self.texcoords_list[v.index].append( uv )
                    face_vert_index.append( int( uv.z ) )
                
            #ǉ_CfbNXlʃXg쐬
            self.face_list.append( face_vert_index )

        #g_܂߂_Xg
        self.vertex_list = []
        for v in mesh.verts:
            self.vertex_list.append( v )
        for vindex in self.extension_verts:
            self.vertex_list.append( mesh.verts[vindex] )



    def outputFrameTransformMatrix( self, mat ):
        """t[so"""
        global indent_level
        output( "FrameTransformMatrix {" )
        indent_level += 1
        outputMatrixParam( mat )
        indent_level -= 1
        output( "}" )


    def outputBoneFrame( self, bone ):
        """{[t[̏o"""
        global indent_level
        output( "Frame %s {" % sanitizeName( bone.name ) )
        indent_level += 1

        #ArmatureBone݂͌̃t[̏ԂƂ͊֌WȂ
        #BONESPACEł͉̂3x3̉]݂̂Ȃ̂ARMATURESPACEvZȂ
        boneMat = Matrix( bone.matrix['ARMATURESPACE'] )
        if bone.parent != None:
            invMat = Matrix( bone.parent.matrix['ARMATURESPACE'] )
            boneMat *= invMat.invert()

        self.outputFrameTransformMatrix( boneMat );
        
        #qBone𒲂ׂ
        if bone.hasChildren():
            for child in bone.children:
                self.outputBoneFrame( child )
    
        indent_level -= 1
        output( "}" );


    
    def outputArmatureHierarchy( self, armature_obj ):
        """Armature烋[g{[̎擾擾Kw\f"""
        global indent_level
        
        for bone in armature_obj.getData().bones.values():
            if bone.parent == None:
                self.outputBoneFrame( bone )


    def outputMeshChank( self, mesh_obj ):
        """Mesh`No͂"""
        global indent_level

        mesh = mesh_obj.getData()

        parent = mesh_obj.getParent()
        if parent != None and parent.getType() == 'Armature':
            #eArmaturełMeshLocalMatrixŎOɒ_ϊ
            #Aj[VĐA{[MeshLocal|Ă܂悤ŌʂBlenderƈقȂĂ܂
            #̂MeshpFrame͍쐬ɒ_𒼐ڕϊĂ
            locMat = mesh_obj.getMatrix('localspace')
            mesh.transform( locMat, True )

        #L_̐􂢏o
        if mesh.hasFaceUV():
            #FaceUV΃Xg͍Ȃ
            self.createVertexList( mesh )

        output( "Mesh %s {" % ( sanitizeName( mesh.name ) ) )
        indent_level += 1
    
        #_
        vertsList = mesh.verts

        if mesh.hasFaceUV():
            #UVLȂu
            vertsList = self.vertex_list
        
        output( "%s;" % ( len( vertsList ) ) )
        for i, v in enumerate( vertsList ):
            ret_chr = ""
            if ( i != len( vertsList ) - 1 ):
                ret_chr = ","
            else:
                ret_chr = ";"

            output( "%f;%f;%f;%s" % ( v.co[0], v.co[1], v.co[2], ret_chr ) )

        
        #
        faceList = mesh.faces
        output( "%s;" % ( len( mesh.faces ) ) )
        for i, face in enumerate( mesh.faces ):
            
            ret_chr = ""
            
            if i != len( mesh.faces ) - 1:
                ret_chr = ";,"
            else:
                ret_chr = ";;"

            if mesh.hasFaceUV():
                #_CfbNXu
                faceV = self.face_list[i]
            else:
                faceV = [v.index for v in face.v]

            face_str = "%i;" % ( len(faceV) )
            
            if IsFlipZ:
                # 0,3,2,1
                face_str += "%i," % ( faceV.pop(0) )
                face_str += ",".join([str(v) for v in reversed(faceV)])
            else:
                # 0,1,2,3
                face_str += ",".join([str(v) for v in faceV])
            face_str += ret_chr
            output( face_str )
    
        #@
        if IsNormals:
            self.outputMeshNormalsChank( mesh )
        #L_
        if IsVertexDuplication:
            self.outputVertexDuplicationIndicesChank( mesh )
        #}eA
        if IsMaterials and mesh.getMaterials():
            self.outputMaterialListChank( mesh )
        #TextureW
        if IsTextureCoords and mesh.hasFaceUV():
            self.outputMeshTextureCoords( mesh )
        #VertexColor
        if IsVertexColors:
            self.outputMeshVertexColors( mesh )
        #SkinMesh
        if IsSkinMesh:
            #eArmature鎞
            parent_obj = mesh_obj.getParent()
            if parent_obj != None and parent_obj.getType() == "Armature":
                self.outputSkinMeshData( mesh_obj, parent_obj )
        
        indent_level -= 1
        output( "}" )


    def outputMeshNormalsChank( self, mesh ):
        """MeshNormal`No͂"""
        global indent_level
        output( "MeshNormals {" )
        indent_level += 1
    
        #_@
        output( "%s;" % ( len( mesh.verts ) ) )
        for i, v in enumerate( mesh.verts ):
            ret_chr = ""
            if ( i != len( mesh.verts ) - 1 ):
                ret_chr = ","
            else:
                ret_chr = ";"
    
            output( "%f;%f;%f;%s" % ( v.no[0], v.no[1], v.no[2], ret_chr ) )
             
        #
        output( "%s;" % ( len( mesh.faces ) ) )
        for i, face in enumerate( mesh.faces ):
            ret_chr = ""
            if i != len( mesh.faces ) - 1:
                ret_chr = ";,"
            else:
                ret_chr = ";;"
    
            face_str = "%i;" % ( len(face.v) )
            faceV = [v.index for v in face.v] #copy
            if IsFlipZ:
                # 0,3,2,1
                face_str += "%i," % ( faceV.pop(0) )
                face_str += ",".join([str(v) for v in reversed(faceV)])
            else:
                # 0,1,2,3
                face_str += ",".join([str(v) for v in faceV])
            face_str += ret_chr
            output( face_str )
        
        indent_level -= 1
        output( "}" )


    def outputMeshTextureCoords( self, mesh ):
        """MeshTextureCoords`No͂"""
        global indent_level
        output( "MeshTextureCoords {" )
        indent_level += 1

        output( "%s;" % ( len( self.vertex_list ) ) )
        for i, v in enumerate( self.vertex_list ):
            ret_chr = ""
            if ( i != len( self.vertex_list ) - 1 ):
                ret_chr = ","
            else:
                ret_chr = ";"

            #_z͑Sg_CfbNX
            extIndex = v.index
            if len(mesh.verts) <= i:
                extIndex = i #AԂ̂͂

            for uv in self.texcoords_list[v.index]:
                if uv.z == extIndex:
                    break
            
            output( "%f;%f;%s" % ( uv.x, -uv.y, ret_chr ) )
        
        indent_level -= 1
        output( "}" )


    def outputTextureChank( self, tex_list ):
        """TextureName`No͂"""
        for mtex in tex_list:
            if mtex:
                tex = mtex.tex
                if tex.getType() == 'Image':
                    #čŏImageTexturêݏo͂
                    output( "TextureFilename {" )
                    output( "\t\"%s\";" % ( Blender.sys.basename( tex.getImage().getFilename() ) ) ); #o͂̓t@Ĉ
                    output( "}" )
                    break


    def outputMaterialChank( self, mat ):
        """Material`No͂"""
        global indent_level
        
        if IsRefMaterial:
            output( "{ %s }" % ( mat.name ) )
        else:
            output( "Material %s {" % ( sanitizeName( mat.name ) ) )
            indent_level += 1
            output( "%f;%f;%f;%f;;" % ( mat.R, mat.G, mat.B, mat.alpha ) ) #ambient
            output( "%f;" % ( mat.spec ) )
            output( "%f;%f;%f;;" % ( mat.specR, mat.specG, mat.specB ) )
            output( "0.000000;0.000000;0.000000;;" )
            tex_list = mat.getTextures()
            if tex_list:
                self.outputTextureChank( tex_list )
                
            indent_level -= 1
            output( "}" )


    def outputMaterialListChank( self, mesh ):
        """MeshMaterialList`No͂"""
        global indent_level
        output( "MeshMaterialList {" )
        indent_level += 1
    
        #gp}eA
        output( "%i;" %( len( mesh.getMaterials() ) ) )
        #ʐ
        output( "%i;" %( len( mesh.faces ) ) )
        for i, face in enumerate( mesh.faces ):
            if i != len( mesh.faces ) - 1:
                output( "%i," % (face.mat) )
            else:
                output( "%i;" % (face.mat) )
    
        for mat in mesh.getMaterials():
            self.outputMaterialChank( mat )
            
        indent_level -= 1
        output( "}" )


    def outputMeshVertexColors( self, mesh ):
        """MeshVertexColors`No͂"""
        #_J[̂߂UV̎̂悤ɒ_͖̂Ɠŷ
        #_CfbNX̒_J[𕽋ωďoƂɂ
        global indent_level

        #_CfbNXYƂ_J[Xg
        vertexColor_list = [ [] for i in range( len( mesh.verts) )]
        for i, face in enumerate( mesh.faces ):
            for j, v in enumerate( face.v ):
                if len(face.col) != 0:
                    vcol = [ face.col[j].r, face.col[j].g, face.col[j].b, face.col[j].a ]
                else:
                    vcol = [ 255, 255, 255, 255] #Ȃ甒蓖Ă
                
                vertexColor_list[v.index].append( vcol )
        
        output( "MeshVertexColors {" )
        indent_level += 1

        #_Pʂ̏o
        vertsList = mesh.verts
        
        if mesh.hasFaceUV():
            #UVLȂu
            vertsList = self.vertex_list
        
        output( "%s;" % ( len( vertsList ) ) )
        for i, v in enumerate( vertsList ):
            ret_chr = ""
            if ( i != len( vertsList ) - 1 ):
                ret_chr = ","
            else:
                ret_chr = ";"

            average = [0,0,0,0] #r,g,b,a
            for vcol in vertexColor_list[v.index]:
                average = [ average[j] + vcol[j] for j in range(len(vcol) )]

            for j, col in enumerate( average ):
                average[j] = ( average[j] / float(len(vertexColor_list[v.index])) ) / 255.0
                
            average[3] = 1.0 #alpha1.0Œ
            
            output( "%i;%f;%f;%f;%f;%s" % ( i, average[0], average[1], average[2], average[3], ret_chr ) )
        
        indent_level -= 1
        output( "}" )

    def outputVertexDuplicationIndicesChank( self, mesh ):
        """Ⴂ̓꒸_f[^CfbNXo͂"""
        global indent_level
        output( "VertexDuplicationIndices {" )
        indent_level += 1

        vertsList = mesh.verts
        if mesh.hasFaceUV():
            #UVLȂu
            vertsList = self.vertex_list

        output( "%i;" % len(vertsList) ) #_
        output( "%i;" % len(mesh.verts) ) #_

        for i, v in enumerate( vertsList ):
            if ( i != len( vertsList ) - 1 ):
                output( "%i," % v.index )
            else:
                output( "%i;" % v.index )
        
        indent_level -= 1
        output( "}" )

    def outputSkinWeightsChank( self, mesh_obj, armature_obj ):
        """XLbṼEFCg`No͂"""
        global indent_level
        mesh = mesh_obj.getData()
        armature = armature_obj.getData()

        for bone in armature.bones.values():
            #{[Ɠ_O[v̒_CfbNXXg𓾂
            try:
                vGroupIndexes = mesh.getVertsFromGroup( bone.name )
            except:
                continue
            
            vIndexList = []
            vWeightList = []
            for vIndex in vGroupIndexes:
                sumWaight = 0
                vIndexWeightList = mesh.getVertexInfluences( vIndex )
                for name, weight in vIndexWeightList:
                    if name == bone.name:
                        currentWaight = weight
                        
                    sumWaight += weight
                    
                currentWaight /= sumWaight
                vIndexList.append( vIndex )
                vWeightList.append( currentWaight )
                
                #ǉ_ɓ_CfbNXΒǉĂ
                for i, vExtIndex in enumerate(self.extension_verts):
                    if vExtIndex == vIndex:
                        vIndexList.append( i + len(mesh.verts) ) #ǉ_CfbNXɕϊăXgɒǉ
                        vWeightList.append( currentWaight )

            ret_chr = ""
            output( "SkinWeights {" )
            indent_level += 1
            output( "\"%s\";" % ( sanitizeName( bone.name ) ) )
            output( "%i;" % ( len(vWeightList) ) )

            #index
            for i, vIndex in enumerate( vIndexList ):
                if ( i != len( vIndexList ) - 1 ):
                    ret_chr = ","
                else:
                    ret_chr = ";"
                output( "%i%s" % ( vIndex, ret_chr ) )

            #weight
            for i, vWeight in enumerate( vWeightList ):
                if ( i != len( vWeightList ) - 1 ):
                    ret_chr = ","
                else:
                    ret_chr = ";"
                output( "%f%s" % ( vWeight, ret_chr ) )

            #BoneOffsetLocal(Armature)W{[Ԃɕϊs
            #ȇ΍s񂩂Ǝvhc{ɂ͂܂
            boneMat = Matrix( bone.matrix['ARMATURESPACE'] )
            offsetMat = boneMat.invert()
            outputMatrixParam( offsetMat );

            indent_level -= 1
            output( "}" )


    def outputSkinMeshData( self, mesh_obj, armature_obj ):
        """XLbVɕKvȃf[^o͂"""
        global indent_level
        mesh = mesh_obj.getData()
        
        maxSkinWeightsPerVertex = 0
        for vGroupName in mesh.getVertGroupNames():
            for vIndex in mesh.getVertsFromGroup( vGroupName ):
                vNum = len( mesh.getVertexInfluences( vIndex ) )
                if vNum > maxSkinWeightsPerVertex:
                    maxSkinWeightsPerVertex = vNum
                
        output( "XSkinMeshHeader {" )
        indent_level += 1
        output( "%i;%i;%i;" % ( maxSkinWeightsPerVertex, maxSkinWeightsPerVertex*3, len( mesh.getVertGroupNames() ) ) )
        indent_level -= 1
        output( "}" )

        self.outputSkinWeightsChank( mesh_obj, armature_obj )


    def outputMeshFrame( self, mesh_obj ):
        """bVt[o͂"""
        global indent_level

        
        parent = mesh_obj.getParent()
        if parent != None and parent.getType() == "Armature" and IsSkinMesh:
            #eArmaturełMesh`No͂
            
            #Mesh`No
            self.outputMeshChank( mesh_obj )
            
            #{[\f
            self.outputArmatureHierarchy( parent )

            #eArmatureꍇqȂ݂Ȃ
            
        else:
            #eArmatureorArmatureŖꍇMeshFrameo͂
            output( "Frame %s {" % ( sanitizeName( mesh_obj.getName() ) ) )
            indent_level += 1
            self.outputFrameTransformMatrix( mesh_obj.getMatrix('localspace') ) #Mesh̃[JXy[X

            #Mesh`No
            self.outputMeshChank( mesh_obj )

            #qIuWFNgf
            child_list = self.getChildren( mesh_obj )
            for child_obj in child_list:
                self.outputObject( child_obj )
    
            indent_level -= 1
            output( "}" )


    def outputAnimetionData( self ):
        """Aj[VɕKvȃf[^ނ̏o"""
        global indent_level
        
        #Animation speed
        output("AnimTicksPerSecond {" )
        indent_level += 1
        output("%i;" % ( AnimTicksPerSecond ) )
        indent_level -= 1
        output("}")

        backup_frame = Blender.Get('curframe')

        #TODO:AnimetionSet̕l

        
        actions = Blender.Armature.NLA.GetActions()
        if len( actions ) > 0:
            #Armature݂Ȃ炻̖OActionƂAnimetipnSet
            for key, action in actions.iteritems():
                output_armobj_list = []
                output("AnimationSet %s {" % ( sanitizeName( key ) ) )
                indent_level += 1
                for obj in self.object_list:
                    parent = obj.getParent()
                    if parent != None and parent.getType() == "Armature" and IsSkinMesh:
                        #{[Aj[Vp
                        if output_armobj_list.count( parent ) == 0: #Armatureo͂Ȃ悤
                            self.outputBoneAnimetionChank( parent, action )
                            output_armobj_list.append( parent )
                    else:
                        #̃Aj[Vp(ArmatureɊ֘AtꂽMeshȊOS)
                        self.outputAnimetionChank( obj )
                    
                indent_level -= 1
                output("}")
        else:
            #Action݂Ȃꍇ
            output("AnimationSet rigid {" )
            indent_level += 1
            
            for obj in self.object_list:
                #̃Aj[Vp(ArmatureɊ֘AtꂽMeshȊOS)
                self.outputAnimetionChank( obj )
                
            indent_level -= 1
            output("}")

        Blender.Set( 'curframe', backup_frame )


    def outputAnimetionChank( self, obj ):
        """ėpIuWFNgpAnimationChank̏o"""
        global indent_level
        frame_list = []
        ipo = obj.getIpo()
        
        if ipo:

            #J[uL[ʒu擾
            for curve in ipo.curves:
                for bezier in curve.bezierPoints:
                    frame = int( bezier.pt[0] )
                    if frame_list.count( frame ) == 0:
                        frame_list.append( frame )
            
            if len( frame_list ) > 0:
                #Object has animation.
                frame_list.sort()
                output("Animation {") 
                indent_level += 1
                output("AnimationKey {" )
                indent_level += 1
                output( "4;" ) #matrix
                output( "%i;" % ( len(frame_list) ) ) 
                for i, frame in enumerate( frame_list ):
                    Blender.Set( 'curframe', frame )

                    ret_chr = ""
                    if i == len(frame_list) - 1:
                        ret_chr = ";"
                    else:
                        ret_chr = ","
                        
                    info = "%i;16;" + (15 * "%f,") + "%f;;" + ret_chr

                    objMat = Matrix( obj.getMatrix('localspace' ) )

                    output(info % (frame,
                        objMat[0][0], objMat[0][1], objMat[0][2], objMat[0][3],
                        objMat[1][0], objMat[1][1], objMat[1][2], objMat[1][3],
                        objMat[2][0], objMat[2][1], objMat[2][2], objMat[2][3],
                        objMat[3][0], objMat[3][1], objMat[3][2], objMat[3][3],
                    ))

                indent_level -= 1
                output( "}" )

                #FrameQ
                output( "{ %s }" % sanitizeName(obj.name) )
                indent_level -= 1
                output( "}" )

        Blender.Set( 'curframe', 0 )

    def getChildFramePoint( self, ipos, bone, frame_list ):
        """̃{[nIPOɑłꂽt[̈ʒu擾"""
        if ipos.has_key( bone.name ):
            ipo = ipos[ bone.name ]
            #J[uL[ʒu擾
            for curve in ipo.curves:
                for bezier in curve.bezierPoints:
                    frame = int( bezier.pt[0] )
                    if frame_list.count( frame ) == 0:
                        frame_list.append( frame )
            
        #q{[𑖍
        for child_bone in bone.children:
            self.getChildFramePoint( ipos, child_bone, frame_list )


    def outputBoneAnimetionChank( self, armature_obj, action ):
        """{[pAnimationChank̏o"""
        global indent_level
        action.setActive( armature_obj )
        armature = armature_obj.getData()
        pose = armature_obj.getPose()
        ipos = action.getAllChannelIpos()

        for bone in armature.bones.values():
            frame_list = []
            if IsParentAnimKey:
                #q{[IPOXg̃L[ׂđ
                self.getChildFramePoint( ipos, bone, frame_list )
            else:
                if ipos.has_key( bone.name ) == False:
                    continue #`l
                ipo = ipos[ bone.name ]
                #J[uL[ʒu擾
                for curve in ipo.curves:
                    for bezier in curve.bezierPoints:
                        frame = int( bezier.pt[0] )
                        if frame_list.count( frame ) == 0:
                            frame_list.append( frame )
            
            if len( frame_list ) > 0:
                #Bone has animation.
                frame_list.sort()
                output("Animation {") 
                indent_level += 1
                output("AnimationKey {" )
                indent_level += 1
                output( "4;" ) #matrix
                output( "%i;" % ( len(frame_list) ) ) 

                for i, frame in enumerate( frame_list ):
                    Blender.Set( 'curframe', frame )

                    ret_chr = ""
                    if i == len(frame_list) - 1:
                        ret_chr = ";"
                    else:
                        ret_chr = ","
                        
                    info = "%i;16;" + (15 * "%f,") + "%f;;" + ret_chr

                    poseBone = pose.bones[bone.name]
                    boneMat = Matrix( poseBone.poseMatrix )
                    if poseBone.parent != None:
                        invBoneMat = Matrix( poseBone.parent.poseMatrix )
                        boneMat *= invBoneMat.invert()

                    output(info % (frame,
                        boneMat[0][0], boneMat[0][1], boneMat[0][2], boneMat[0][3],
                        boneMat[1][0], boneMat[1][1], boneMat[1][2], boneMat[1][3],
                        boneMat[2][0], boneMat[2][1], boneMat[2][2], boneMat[2][3],
                        boneMat[3][0], boneMat[3][1], boneMat[3][2], boneMat[3][3],
                    ))

                indent_level -= 1
                output( "}" )

                #FrameQ
                output( "{ %s }" % sanitizeName(bone.name) )
                indent_level -= 1
                output( "}" )
                

        Blender.Set( 'curframe', 0 )


    def outputObject( self, obj ):
        """eIuWFNgo"""
        if obj.getType() == "Armature":
            self.outputObjectFrame( obj )
        if obj.getType() == "Empty":
            self.outputObjectFrame( obj )
        if obj.getType() == "Mesh":
            self.outputMeshFrame( obj )
        
        self.object_list.append( obj )


    def outputObjectFrame( self, obj ):
        """ėp(ArmatureEmpty)Framȅo"""
        global indent_level
        output( "Frame %s {" % sanitizeName( obj.getName() ) )
        indent_level += 1

        self.outputFrameTransformMatrix( obj.getMatrix('localspace') ) #Object̃[JXy[X

        #qIuWFNgf
        child_list = self.getChildren( obj )
        for child_obj in child_list:
            self.outputObject( child_obj )

        indent_level -= 1
        output( "}" )


    def outputTemplateDefinition( self ):
        """DirectX File Templateo"""
        output("template VertexDuplicationIndices { \n\
 <b8d65549-d7c9-4995-89cf-53a9a8b031e3>\n\
 DWORD nIndices;\n\
 DWORD nOriginalVertices;\n\
 array DWORD indices[nIndices];\n\
}\n\
template XSkinMeshHeader {\n\
 <3cf169ce-ff7c-44ab-93c0-f78f62d172e2>\n\
 WORD nMaxSkinWeightsPerVertex;\n\
 WORD nMaxSkinWeightsPerFace;\n\
 WORD nBones;\n\
}\n\
template SkinWeights {\n\
 <6f0d123b-bad2-4167-a0d0-80224f25fabb>\n\
 STRING transformNodeName;\n\
 DWORD nWeights;\n\
 array DWORD vertexIndices[nWeights];\n\
 array float weights[nWeights];\n\
 Matrix4x4 matrixOffset;\n\
}\n\
template AnimTicksPerSecond {\n\
 <9E415A43-7BA6-4a73-8743-B73D47E88476>\n\
 DWORD AnimTicksPerSecond;\n\
}\n\n")
        
    def startExport( self ):
        """GNX|[gJn"""
        global indent_level
        #wb_
        output( "xof 0303txt 0032\n\n" )
        
        self.outputTemplateDefinition()
        
        #[gt[Ă
        output( "Frame RootFrame {" )
        indent_level += 1
        mat = Matrix().identity()
        
        if IsSwapYZ:
            mat *= rotMat
        if IsFlipZ:
            mat *= flipMat
        
        self.outputFrameTransformMatrix( mat ) # [gȂPʍs

        self.object_list = []

        if IsOutputAll:
            #SIuWFNgo
            for obj in Blender.Scene.GetCurrent().getChildren():
                if obj.getParent() == None:
                    #e̋ȂIuWFNĝݏo
                    self.outputObject( obj )
        else:
            #IIuWFNgƂ̎q̂ݏo
            for obj in Blender.Object.GetSelected():
                if self.object_list.count( obj ) == 0:
                    self.outputObject( obj )
        
        indent_level -= 1
        output( "}" )

        #ŌɃAj[V֘A
        if IsAnimetion:
            self.outputAnimetionData()

#===========GUI===========

def selection_callback(filename):
    global contents
    """Blender̃t@CIWindowĂ΂R[obN"""
    if filename.find('.x', -2) <= 0: filename += '.x' 

    contents = []
    
    xfile = XFileExporter()
    xfile.startExport()
    f = file( filename, "w" )
    f.write( "".join( contents ) )
    f.close()

def OnButton(event):
    """{^Cxg"""
    global IsOutputAll, IsSwapYZ, IsFlipZ, IsAnimetion, AnimTicksPerSecond
    global IsNormals, IsMaterials, IsRefMaterial, IsTextureCoords, IsVertexColors, IsSkinMesh

    if event in (1, 2, 3, 4, 5, 6, 7, 8, 10):
        if event == 1:
            IsSwapYZ = not IsSwapYZ
        elif event == 2:
            IsFlipZ = not IsFlipZ
        elif event == 3:
            IsAnimetion = not IsAnimetion
        elif event == 4:
            IsNormals = not IsNormals
        elif event == 5:
            IsVertexColors = not IsVertexColors
        elif event == 6:
            IsVertexDuplication = not IsVertexDuplication
        elif event == 7:
            IsMaterials = not IsMaterials
        elif event == 8:
            IsTextureCoords = not IsTextureCoords
        elif event == 9:
            IsSkinMesh = not IsSkinMesh
        elif event == 10:
            AnimTicksPerSecond = AnimTickCtrl.val

        Draw.Redraw(1)
    elif event in (100, 101):
        if event == 100:
            IsOutputAll = True
        elif event == 101:
            IsOutputAll = False

        name = Blender.Get('filename').strip(".blend") + ".x"
        Blender.Window.FileSelector( selection_callback, "Export", name )
    elif event == 102:
        Blender.Draw.Exit()  


def OnEvent(event, value):
    """L[Cxg"""
    if event == Blender.Draw.ESCKEY:
        Blender.Draw.Exit()   


def OnDraw():
    """GUI`ݒ"""
    global AnimTicksPerSecond, AnimTickCtrl

    TOP = 300
    LEFT = 30
    RIGHT = 150
    EDATA = TOP - 110 
    MISC = EDATA - 30
    DOIT = MISC - 110
    
    glClear(GL_COLOR_BUFFER_BIT)
    glColor3f(0.1, 0.1, 0.1)
    
    glRasterPos2i(LEFT, TOP+30)
    Draw.Text("Export Option", "large")
    Draw.Toggle("SwapYZ",    1,  LEFT,    TOP, 95, 20, IsSwapYZ, "swap up axis Y or Z.")
    Draw.Toggle("FlipZ",     2, RIGHT,    TOP, 95, 20, IsFlipZ, "change rihght to left hand system.")
    Draw.Toggle("Animation", 3,  LEFT, TOP-30, 95, 20, IsAnimetion, "Export Animetion.")
    AnimTickCtrl = Draw.Number("Speed", 10, RIGHT, TOP-30, 95, 20, AnimTicksPerSecond, 1, 1000, "tick par second") 
    Draw.Toggle("Parent IPO Key", 9,  LEFT, TOP-60, 95, 20, IsParentAnimKey, "Export parent Bone IPO Keys.")   
    
    glRasterPos2i(LEFT, EDATA+30)
    Draw.Text("Exported chank data", "large")
    Draw.Toggle("Normals",      4,  LEFT,    EDATA, 95, 20, IsNormals, "Export vertex normals.")
    Draw.Toggle("VertexColors", 5, RIGHT,    EDATA, 95, 20, IsVertexColors, "Export vertex diffuse colors.")
    Draw.Toggle("VertexDuplication",    6,  LEFT, EDATA-30, 95, 20, IsVertexDuplication, "Export vertex duplication indices.") 
    Draw.Toggle("Materials",    7,  LEFT, EDATA-60, 95, 20, IsMaterials, "Export materials.") 
    Draw.Toggle("Texcoords",    8,  LEFT, EDATA-90, 95, 20, IsTextureCoords, "Export vertex texture coordinates.")   
    Draw.Toggle("SkinMesh",     9, RIGHT, EDATA-90, 95, 20, IsSkinMesh, "Export wait map.")   
    
    Draw.Button("ExportAll",    100,     LEFT, DOIT, 80, 30, "Export data.")
    Draw.Button("ExportSelect", 101,  LEFT+90, DOIT, 80, 30, "Export selection data.")
    Draw.Button("Exit",         102, LEFT+180, DOIT, 80, 30, "Exit the script.")
     
    glColor3f(0.3, 0.3, 0.3)
    glRasterPos2i(LEFT, 35)
    Draw.Text("(C) 2007 kakkou (DirectX File Exporter Ver:0.8)", "small")

#==============================
if 0:
    #foOp
    xfile = XFileExporter()
    xfile.startExport()
    f = file( "c:\hoge.x", "w" )
    f.write( "".join( contents ) )
    f.close()
else:
    Blender.Draw.Register(OnDraw, OnEvent, OnButton)